from _base_types cimport BaseType, CharType, SizeType, UCharType, ValueType
from _guide_unit cimport GuideUnit
from iostream cimport istream, ostream


cdef extern from "../lib/dawgdic/src/dawgdic/guide.h" namespace "dawgdic":
    cdef cppclass Guide:

        Guide()

        GuideUnit *units()
        SizeType size()
        SizeType total_size()
        SizeType file_size()

        # The root index.
        BaseType root()

        UCharType child(BaseType index)
        UCharType sibling(BaseType index)

        # Reads a dictionary from an input stream.
        bint Read(istream *input)

        # Writes a dictionry to an output stream.
        bint Write(ostream *output)

        # Maps memory with its size.
        void Map(void *address)

        # Swaps Guides.
        void Swap(Guide *Guide)

        # Initializes a Guide.
        void Clear()


from _base_types cimport BaseType, CharType, SizeType, UCharType, ValueType
from _dictionary_unit cimport DictionaryUnit
from iostream cimport istream, ostream


cdef extern from "../lib/dawgdic/src/dawgdic/dictionary.h" namespace "dawgdic":
    cdef cppclass Dictionary:

        Dictionary() nogil

        DictionaryUnit *units() nogil
        SizeType size() nogil
        SizeType total_size() nogil
        SizeType file_size() nogil

        # Root index.
        BaseType root() nogil

        # Checks if a given index is related to the end of a key.
        bint has_value(BaseType index) nogil

        # Gets a value from a given index.
        ValueType value(BaseType index) nogil

        # Reads a dictionary from an input stream.
        bint Read(istream *input) nogil except +

        # Writes a dictionry to an output stream.
        bint Write(ostream *output) nogil except +

        # Exact matching.
        bint Contains(CharType *key) nogil
        bint Contains(CharType *key, SizeType length) nogil

        # Exact matching.
        ValueType Find(CharType *key) nogil
        ValueType Find(CharType *key, SizeType length) nogil
        bint Find(CharType *key, ValueType *value) nogil
        bint Find(CharType *key, SizeType length, ValueType *value) nogil

        # Follows a transition.
        bint Follow(CharType label, BaseType *index) nogil

        # Follows transitions.
        bint Follow(CharType *s, BaseType *index) nogil
        bint Follow(CharType *s, BaseType *index, SizeType *count) nogil

        # Follows transitions.
        bint Follow(CharType *s, SizeType length, BaseType *index) nogil
        bint Follow(CharType *s, SizeType length, BaseType *index, SizeType *count) nogil

        # Maps memory with its size.
        void Map(void *address) nogil
        void Map(void *address, SizeType size) nogil

        # Initializes a dictionary.
        void Clear() nogil

        # Swaps dictionaries.
        void Swap(Dictionary *dic) nogil
        # Shrinks a vector.
        void Shrink() nogil


cdef extern from "../lib/dawgdic/src/dawgdic/base-types.h" namespace "dawgdic":
    # 8-bit characters.
    ctypedef char CharType
    ctypedef unsigned char UCharType

    # 32-bit integer.
    ctypedef int ValueType

    # 32-bit unsigned integer.
    ctypedef unsigned int BaseType

    # 32 or 64-bit unsigned integer.
    ctypedef int SizeType




import pytest

import dawg


class TestBytesDAWG:

    DATA = (
        ("foo", b"data3"),
        ("bar", b"data2"),
        ("foo", b"data1"),
        ("foobar", b"data4"),
    )

    DATA_KEYS = list(zip(*DATA))[0]

    def dawg(self, **kwargs):
        return dawg.BytesDAWG(self.DATA, **kwargs)

    def test_contains(self):
        d = self.dawg()
        for key, val in self.DATA:
            assert key in d

        assert "food" not in d
        assert "x" not in d
        assert "fo" not in d

    def test_getitem(self):
        d = self.dawg()

        assert d["foo"] == [b"data1", b"data3"]
        assert d["bar"] == [b"data2"]
        assert d["foobar"] == [b"data4"]

        with pytest.raises(KeyError):
            d["f"]

        with pytest.raises(KeyError):
            d["food"]

        with pytest.raises(KeyError):
            d["foobarz"]

        with pytest.raises(KeyError):
            d["x"]

    def test_prefixes(self):
        d = self.dawg()
        assert d.prefixes("foobarz") == ["foo", "foobar"]
        assert d.prefixes("x") == []
        assert d.prefixes("bar") == ["bar"]

    def test_keys(self):
        d = self.dawg()
        assert d.keys() == sorted(self.DATA_KEYS)

    def test_keys_ordering(self):
        data = [("foo", b"v1"), ("foobar", b"v2"), ("bar", b"v3")]

        d = dawg.BytesDAWG(data, payload_separator=b"\xff")
        assert d.keys() == ["bar", "foobar", "foo"]

        d2 = dawg.BytesDAWG(data, payload_separator=b"\x01")
        assert d2.keys() == ["bar", "foo", "foobar"]

    def test_iterkeys(self):
        d = self.dawg()
        assert list(d.iterkeys()) == d.keys()
        assert list(d.iterkeys()) == sorted(self.DATA_KEYS)

    def test_items(self):
        d = self.dawg()
        assert d.items() == sorted(self.DATA)

    def test_iteritems(self):
        d = self.dawg()
        assert list(d.iteritems()) == d.items()

    def test_build_error(self):
        with pytest.raises(dawg.Error):
            self.dawg(payload_separator=b"f")


class TestRecordDAWG:

    STRUCTURED_DATA = (
        ("foo", (3, 2, 256)),
        ("bar", (3, 1, 0)),
        ("foo", (3, 2, 1)),
        ("foobar", (6, 3, 0)),
    )

    def dawg(self):
        return dawg.RecordDAWG(">3H", self.STRUCTURED_DATA)

    def test_record_getitem(self):
        d = self.dawg()
        assert d["foo"] == [(3, 2, 1), (3, 2, 256)]
        assert d["bar"] == [(3, 1, 0)]
        assert d["foobar"] == [(6, 3, 0)]

    def test_record_items(self):
        d = self.dawg()
        assert d.items() == sorted(self.STRUCTURED_DATA)

    def test_record_keys(self):
        d = self.dawg()
        assert d.keys() == [
            "bar",
            "foo",
            "foo",
            "foobar",
        ]

    def test_record_iterkeys(self):
        d = self.dawg()
        assert list(d.iterkeys()) == d.keys()

    def test_record_iteritems(self):
        d = self.dawg()
        assert list(d.iteritems()) == d.items()

    def test_record_keys_prefix(self):
        d = self.dawg()
        assert d.keys("fo") == ["foo", "foo", "foobar"]
        assert d.keys("bar") == ["bar"]
        assert d.keys("barz") == []

    def test_prefixes(self):
        d = self.dawg()
        assert d.prefixes("foobarz") == ["foo", "foobar"]
        assert d.prefixes("x") == []
        assert d.prefixes("bar") == ["bar"]



import pickle
import tempfile
from io import BytesIO

import pytest

import dawg


def test_contains():
    d = dawg.IntDAWG({"foo": 1, "bar": 2, "foobar": 3})

    assert "foo" in d
    assert "bar" in d
    assert "foobar" in d
    assert "fo" not in d
    assert "x" not in d

    assert b"foo" in d
    assert b"x" not in d


class TestDAWG:
    def test_sorted_iterable(self):

        sorted_data = ["bar", "foo", "foobar"]
        contents = "\n".join(sorted_data).encode("utf8")
        with tempfile.NamedTemporaryFile() as f:
            f.write(contents)
            f.seek(0)

            words = (line.strip() for line in f)
            d = dawg.DAWG(words, input_is_sorted=True)

        assert "bar" in d
        assert "foo" in d

    def test_no_segfaults_on_invalid_file(self):
        d = dawg.DAWG()
        fd, path = tempfile.mkstemp()
        with open(path, "w") as f:
            f.write("foo")

        with pytest.raises(IOError) as e:
            d.load(path)
            assert "Invalid" in e.args[0]

        with open(path, "rb") as f:
            with pytest.raises(IOError) as e:
                d.read(f)
                assert "Invalid" in e.args[0]

    def test_no_segfaults_after_wrong_stream(self):
        d = dawg.DAWG()
        wrong_path = tempfile.mktemp()  # file doesn't exists

        with pytest.raises(IOError):
            d.load(wrong_path)

        assert "random-key" not in d  # there is possible segfault

    def test_build_errors(self):
        with pytest.raises(dawg.Error):
            data = [b"foo\x00bar", b"bar"]
            dawg.DAWG(data)

    def test_contains_with_null_bytes(self):
        d = dawg.DAWG(["foo"])
        assert b"foo" in d
        assert b"foo\x00bar" not in d

    def test_unicode_sorting(self):
        key1 = "\U00010345\U0001033f\U00010337\U00010330\U0001033d"
        key2 = "\uff72\uff9c\uff90\uff7b\uff9e\uff9c"

        # This apparently depends on Python version:
        # assert key1 < key2
        # assert key1.encode('utf8') > key2.encode('utf8')

        # Constructor should sort data according to utf8 values,
        # not according to unicode sorting rules. It will raise an exception
        # if data is sorted according to unicode rules.
        dawg.DAWG([key1, key2])


class TestIntDAWG:

    IntDAWG = dawg.IntDAWG

    def dawg(self):
        payload = {"foo": 1, "bar": 5, "foobar": 3}
        d = self.IntDAWG(payload)
        return payload, d

    def test_getitem(self):
        payload, d = self.dawg()
        for key in payload:
            assert d[key] == payload[key]

        with pytest.raises(KeyError):
            d["fo"]

    def test_dumps_loads(self):
        payload, d = self.dawg()
        data = d.tobytes()

        d2 = self.IntDAWG()
        d2.frombytes(data)
        for key, value in payload.items():
            assert key in d2
            assert d2[key] == value

    def test_dump_load(self):
        payload, _ = self.dawg()

        buf = BytesIO()
        self.IntDAWG(payload).write(buf)
        buf.seek(0)

        d = self.IntDAWG()
        d.read(buf)

        for key, value in payload.items():
            assert key in d
            assert d[key] == value

    def test_pickling(self):
        payload, d = self.dawg()

        data = pickle.dumps(d)
        d2 = pickle.loads(data)

        for key, value in payload.items():
            assert key in d2
            assert d[key] == value

    def test_int_value_ranges(self):
        for val in [0, 5, 2**16 - 1, 2**31 - 1]:
            d = self.IntDAWG({"f": val})
            assert d["f"] == val

        with pytest.raises(ValueError):
            self.IntDAWG({"f": -1})

        with pytest.raises(OverflowError):
            self.IntDAWG({"f": 2**32 - 1})


class TestIntCompletionDAWG(TestIntDAWG):
    IntDAWG = dawg.IntCompletionDAWG  # checks that all tests for IntDAWG pass


class TestCompletionDAWG:
    keys = ["f", "bar", "foo", "foobar"]

    def dawg(self):
        return dawg.CompletionDAWG(self.keys)

    def empty_dawg(self):
        return dawg.CompletionDAWG()

    def test_contains(self):
        d = self.dawg()
        for key in self.keys:
            assert key in d

    def test_keys(self):
        d = self.dawg()
        assert d.keys() == sorted(self.keys)

    def test_iterkeys(self):
        d = self.dawg()
        assert list(d.iterkeys()) == sorted(self.keys)
        assert list(d.iterkeys()) == d.keys()

    def test_prefixes(self):
        d = self.dawg()
        assert d.prefixes("foobarz") == ["f", "foo", "foobar"]
        assert d.prefixes("x") == []
        assert d.prefixes("bar") == ["bar"]

    def test_b_prefixes(self):
        d = self.dawg()
        assert d.b_prefixes(b"foobarz") == [b"f", b"foo", b"foobar"]
        assert d.b_prefixes(b"x") == []
        assert d.b_prefixes(b"bar") == [b"bar"]

    def test_iterprefixes(self):
        d = self.dawg()
        assert list(d.iterprefixes("foobarz")) == d.prefixes("foobarz")
        assert list(d.iterprefixes("x")) == d.prefixes("x")
        assert list(d.iterprefixes("bar")) == d.prefixes("bar")

    def test_completion(self):
        d = self.dawg()

        assert d.keys("z") == []
        assert d.keys("b") == ["bar"]
        assert d.keys("foo") == ["foo", "foobar"]

    def test_has_keys_with_prefix(self):
        assert self.empty_dawg().has_keys_with_prefix("") is False

        d = self.dawg()
        assert d.has_keys_with_prefix("") is True
        assert d.has_keys_with_prefix("b") is True
        assert d.has_keys_with_prefix("fo") is True
        assert d.has_keys_with_prefix("bo") is False

    def test_completion_dawg_saveload(self):
        buf = BytesIO()
        self.dawg().write(buf)
        buf.seek(0)

        d = self.empty_dawg()
        d.read(buf)

        for key in self.keys:
            assert key in d

        assert d.keys("foo") == ["foo", "foobar"]
        assert d.keys("b") == ["bar"]
        assert d.keys("z") == []

    def test_no_segfaults_on_invalid_file(self):
        d = self.dawg()
        fd, path = tempfile.mkstemp()
        with open(path, "w") as f:
            f.write("foo")

        with pytest.raises(IOError) as e:
            d.load(path)
            assert "can't load _dawg.Dictionary" in e.args[0]

    def test_no_segfaults_on_empty_dawg(self):
        d = dawg.CompletionDAWG([])
        assert d.keys() == []


class TestIntCompletionDAWGComplete(TestCompletionDAWG):
    keys = ["f", "bar", "foo", "foobar"]

    def dawg(self):
        return dawg.IntCompletionDAWG((k, len(k)) for k in self.keys)

    def empty_dawg(self):
        return dawg.IntCompletionDAWG()

    def test_no_segfaults_on_empty_dawg(self):
        d = dawg.IntCompletionDAWG([])
        assert d.keys() == []

    def test_items(self):
        d = self.dawg()
        items = d.items()
        assert isinstance(items, list)
        for key, value in items:
            assert len(key) == value

    def test_iteritems(self):
        d = self.dawg()
        for key, value in d.iteritems():
            assert len(key) == value

    def test_items_prefix(self):
        d = self.dawg()
        assert d.items("fo") == [("foo", 3), ("foobar", 6)]



from _base_types cimport BaseType, CharType, SizeType, UCharType, ValueType


cdef extern from "../lib/dawgdic/src/dawgdic/dictionary-unit.h" namespace "dawgdic":
    cdef cppclass DictionaryUnit:

        DictionaryUnit() nogil

        # Sets a flag to show that a unit has a leaf as a child.
        void set_has_leaf() nogil

        # Sets a value to a leaf unit.
        void set_value(ValueType value) nogil

        # Sets a label to a non-leaf unit.
        void set_label(UCharType label) nogil

        # Sets an offset to a non-leaf unit.
        bint set_offset(BaseType offset) nogil


        # Checks if a unit has a leaf as a child or not.
        bint has_leaf() nogil

        # Checks if a unit corresponds to a leaf or not.
        ValueType value() nogil

        # Reads a label with a leaf flag from a non-leaf unit.
        BaseType label() nogil

        # Reads an offset to child units from a non-leaf unit.
        BaseType offset() nogil



import pytest

import dawg


class TestPrediction:
    DATA = [
        "ЁЖИК",
        "ЁЖИКЕ",
        "ЁЖ",
        "ДЕРЕВНЯ",
        "ДЕРЁВНЯ",
        "ЕМ",
        "ОЗЕРА",
        "ОЗЁРА",
        "ОЗЕРО",
    ]
    LENGTH_DATA = list(zip(DATA, ((len(w),) for w in DATA)))

    REPLACES = dawg.DAWG.compile_replaces({"Е": "Ё"})

    SUITE = [
        ("УЖ", []),
        ("ЕМ", ["ЕМ"]),
        ("ЁМ", []),
        ("ЁЖ", ["ЁЖ"]),
        ("ЕЖ", ["ЁЖ"]),
        ("ЁЖИК", ["ЁЖИК"]),
        ("ЕЖИКЕ", ["ЁЖИКЕ"]),
        ("ДЕРЕВНЯ", ["ДЕРЕВНЯ", "ДЕРЁВНЯ"]),
        ("ДЕРЁВНЯ", ["ДЕРЁВНЯ"]),
        ("ОЗЕРА", ["ОЗЕРА", "ОЗЁРА"]),
        ("ОЗЕРО", ["ОЗЕРО"]),
    ]

    SUITE_ITEMS = [(it[0], [(w, [(len(w),)]) for w in it[1]]) for it in SUITE]  # key  # item, value pair

    SUITE_VALUES = [(it[0], [[(len(w),)] for w in it[1]]) for it in SUITE]  # key

    @pytest.mark.parametrize(("word", "prediction"), SUITE)
    def test_dawg_prediction(self, word, prediction):
        d = dawg.DAWG(self.DATA)
        assert d.similar_keys(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE)
    def test_record_dawg_prediction(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_keys(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE_ITEMS)
    def test_record_dawg_items(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_items(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE_VALUES)
    def test_record_dawg_items_values(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_item_values(word, self.REPLACES) == prediction


class TestMultiValuedPrediction:
    DATA = "хлѣб ёлка ель лѣс лѣсное всё всѣ бѣлёная изобрѣтён лев лёв лѣв вѣнскій".split(" ")
    LENGTH_DATA = list(zip(DATA, ((len(w),) for w in DATA)))

    REPLACES = dawg.DAWG.compile_replaces({"е": ["ё", "ѣ"], "и": "і"})

    SUITE = [
        ("осел", []),
        ("ель", ["ель"]),
        ("ёль", []),
        ("хлеб", ["хлѣб"]),
        ("елка", ["ёлка"]),
        ("лесное", ["лѣсное"]),
        ("лесноё", []),
        ("лёсное", []),
        ("изобретен", ["изобрѣтён"]),
        ("беленая", ["бѣлёная"]),
        ("белёная", ["бѣлёная"]),
        ("бѣленая", ["бѣлёная"]),
        ("бѣлёная", ["бѣлёная"]),
        ("белѣная", []),
        ("бѣлѣная", []),
        ("все", ["всё", "всѣ"]),
        ("лев", ["лев", "лёв", "лѣв"]),
        ("венский", ["вѣнскій"]),
    ]

    SUITE_ITEMS = [(it[0], [(w, [(len(w),)]) for w in it[1]]) for it in SUITE]  # key  # item, value pair

    SUITE_VALUES = [(it[0], [[(len(w),)] for w in it[1]]) for it in SUITE]  # key

    @pytest.mark.parametrize(("word", "prediction"), SUITE)
    def test_dawg_prediction(self, word, prediction):
        d = dawg.DAWG(self.DATA)
        assert d.similar_keys(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE)
    def test_record_dawg_prediction(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_keys(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE_ITEMS)
    def test_record_dawg_items(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_items(word, self.REPLACES) == prediction

    @pytest.mark.parametrize(("word", "prediction"), SUITE_VALUES)
    def test_record_dawg_items_values(self, word, prediction):
        d = dawg.RecordDAWG("=H", self.LENGTH_DATA)
        assert d.similar_item_values(word, self.REPLACES) == prediction



#! /usr/bin/env python
import glob
import os

from Cython.Build import cythonize
from setuptools import Extension, setup

TEST = os.environ.get("TEST", "false").casefold() == "true"

compiler_directives = {"language_level": 3}
define_macros = []

if TEST:
    compiler_directives.update({"linetrace": True, "profile": True})
    define_macros.extend([("CYTHON_TRACE", "1"), ("CYTHON_TRACE_NOGIL", "1")])

extensions = [
    Extension(
        "dawg",
        sources=glob.glob("src/*.pyx") + glob.glob("lib/b64/*.c"),
        include_dirs=["lib"],
        language="c++",
        define_macros=define_macros,
    )
]

ext_modules = cythonize(
    extensions,
    language="c++",
    annotate=False,
    compiler_directives=compiler_directives,
)

setup(
    name="DAWG2",
    version="0.12.1",
    description="Fast and memory efficient DAWG (DAFSA) for Python",
    long_description=open("README.md").read() + "\n\n" + open("CHANGES.md").read(),
    long_description_content_type='text/markdown',
    author="Mikhail Korobov",
    author_email="kmike84@gmail.com",
    url="https://github.com/pymorphy2-fork/DAWG/",
    ext_modules=ext_modules,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing :: Linguistic",
    ],
)
